//#![feature(duration_float)]
#[macro_use]
extern crate structopt;

const SOL_PATH: &str = "../d.cpp";
const TGEN_PATH: &str = "../d.testgen.cpp";
const SEED: &str = "228";
use rand::prelude::*;
use std::{
      cmp::Ordering,
      collections::BTreeMap,
      fs,
      io::{self, Write},
      process::Command,
};
fn build() {
      let sol = Command::new("g++")
            .args(&[
                  SOL_PATH,
                  "-o",
                  "data/mo",
                  "-std=c++1z",
                  "-O2",
                  "-DFROM_FILE",
                  "-DTUNE",
                  "-DSILENT",
                  "-save-temps=obj",
            ])
            .status()
            .unwrap();
      assert!(sol.success());
      let tgen = Command::new("g++")
            .args(&[TGEN_PATH, "-o", "data/tg", "-std=c++1z", "-DSILENT"])
            .status()
            .unwrap();
      assert!(tgen.success());
}

struct Cache {
      cache: std::collections::BTreeMap<(isize, isize), f64>,
}

#[derive(Debug, StructOpt)]
#[structopt(name = "tuner")]
struct Opts {
      #[structopt(short = "m")]
      tamb_mo: isize,
      #[structopt(short = "s")]
      tamb_sqrt: isize,
}

fn singlerun(tamb_mo: isize, tamb_sqrt: isize, cache: &mut Cache) -> f64 {
      let cache_key = (tamb_mo, tamb_sqrt);
      if cache.cache.contains_key(&cache_key) {
            return cache.cache[&cache_key];
      }
      print!("checking MO={}, SQRT={}...", tamb_mo, tamb_sqrt);
      io::stdout().flush().unwrap();
      let tune_options = format!("{} {}\n", tamb_mo, tamb_sqrt);
      fs::write("./data/tune.txt", tune_options).unwrap();
      Command::new("./tg")
            .current_dir("./data")
            .arg(SEED)
            .status()
            .unwrap();
      let time_cur = std::time::Instant::now();
      let res = Command::new("./mo").current_dir("./data").status().unwrap();
      let time_end = std::time::Instant::now();
      assert!(res.success());
      let dur = time_end.duration_since(time_cur);
      let ans = dur.as_secs() as f64 + (dur.subsec_nanos() as f64 * 1e-9);
      cache.cache.insert(cache_key, ans);
      println!("{}", ans);
      ans
}

fn multirun(tamb_mo: isize, tamb_sqrt: isize, cr: &mut Cache) -> f64 {
      let mut times = Vec::new();
      for _i in 0..5 {
            let t = singlerun(tamb_mo, tamb_sqrt, cr);
            times.push(t);
      }
      times.sort_by(|a, b| {
            let diff = a - b;
            if diff > 0.0 {
                  Ordering::Greater
            } else if diff < 0.0 {
                  Ordering::Less
            } else {
                  Ordering::Equal
            }
      });
      let ans = times[0];
      cr.cache.insert((tamb_mo, tamb_sqrt), ans);
      ans
}
//const TAMB_MO: isize = 2000;
//const TAMB_SQRT: isize = 700;
fn main() {
      use structopt::StructOpt;
      let opts: Opts = Opts::from_args();
      build();
      let mut cache = Cache {
            cache: BTreeMap::new(),
      };

      let cr = &mut cache;
      let mut tamb_mo = opts.tamb_mo;
      let mut tamb_sqrt = opts.tamb_sqrt;
      /*for k in [-1 as isize, 1 as isize].iter() {
            while run(tamb_mo + 50 * k, tamb_sqrt, cr) < run(tamb_mo, tamb_sqrt, cr) {
                  tamb_mo += 50 * k;
            }
            while run(tamb_mo, tamb_sqrt + 5 * k, cr) < run(tamb_mo, tamb_sqrt, cr) {
                  tamb_sqrt += 5 * k;
            }
      }*/
      let mut rnd = rand::thread_rng();
      let dist = rand::distributions::Normal::new(0.0, 15.0);
      let initial_temp = 1;
      for i in 0..10 {
            let cur_temp = (initial_temp as f64) / (i as f64 + 1.0);
            let nmo = tamb_mo + dist.sample(&mut rnd).round() as isize;
            let nsqrt = tamb_sqrt + dist.sample(&mut rnd).round() as isize;
            let nres = multirun(nmo, nsqrt, cr);
            let de = nres - multirun(tamb_mo, tamb_sqrt, cr);
            if de < 0.0 || (-de / cur_temp).exp() > rnd.gen() {
                  tamb_mo = nmo;
                  tamb_sqrt = nsqrt;
                  println!("relaxing, dE={}", de);
            }
      }

      let mut best = (0, 0);
      let mut bestsc = 228.0;
      for (k, v) in cache.cache {
            if v < bestsc {
                  best = k;
                  bestsc = v;
            }
      }
      println!(
            "best result: {}, with MO={}, SQRT={}",
            bestsc, best.0, best.1
      );
}
